import gql from "graphql-tag";
import { PbElement, PbErrorResponse, PbPageRevision } from "~/types";

const error = `
    error {
        code
        message
        data
}`;

export interface PageResponseData {
    id: string;
    pid: string;
    title: string;
    path: string;
    version: string;
    locked: boolean;
    status: string;
    revisions: PbPageRevision[];
}
export const DATA_FIELDS = `
    id
    pid
    title
    path
    version
    locked
    status
    revisions {
        id
        savedOn
        locked
        title
        status
        version
    }
    
`;

export const LIST_PAGES_DATA_FIELDS = `
    id
    pid
    status
    title
    version
    savedOn
    category {
        name
        slug
    }
    createdBy {
        id
        displayName
    }
`;

export const CREATE_PAGE = gql`
    mutation PbCreatePage($from: ID, $category: String) {
        pageBuilder {
            createPage(from: $from, category: $category) {
                data {
                    ${LIST_PAGES_DATA_FIELDS}
                }
                ${error}
            }
        }
    }
`;

export const LIST_PAGES = gql`
    query PbListPages(
        $where: PbListPagesWhereInput
        $sort: [PbListPagesSort!]
        $search: PbListPagesSearchInput
        $limit: Int
        $after: String
    ) {
        pageBuilder {
            listPages(where: $where, sort: $sort, limit: $limit, after: $after, search: $search) {
                data {
                    ${LIST_PAGES_DATA_FIELDS}
                }
                meta {
                    totalCount
                    hasMoreItems
                    cursor
                }
                error {
                    data
                    code
                    message
                }
            }
        }
    }
`;
/**
 * ##############################
 * Get Page Query Response
 */
export interface GetPageQueryResponse {
    pageBuilder: {
        getPage: {
            data: PageResponseData | null;
            error: PbErrorResponse | null;
        };
    };
}
export interface GetPageQueryVariables {
    id: string;
}
export const GET_PAGE = gql`
    query PbGetPagePreview($id: ID!) {
        pageBuilder {
            getPage(id: $id) {
                data {
                    ${DATA_FIELDS}
                    createdBy {
                        id
                    }
                    content

                }
                ${error}
            }
        }
    }
`;
export const PUBLISH_PAGE = gql`
    mutation PbPublishPage($id: ID!) {
        pageBuilder {
            publishPage(id: $id) {
                data {
                    ${DATA_FIELDS}
                }
                ${error}
            }
        }
    }
`;

export const UNPUBLISH_PAGE = gql`
    mutation PbUnpublishPage($id: ID!) {
        pageBuilder {
            unpublishPage(id: $id) {
                data {
                    ${DATA_FIELDS}
                }
                ${error}
            }
        }
    }
`;
/**
 * ##########################
 * Delete Page Mutation
 */
interface DeletePageMutationResponseData {
    latestPage: {
        id: string;
        status: string;
        version: number;
    };
}
export interface DeletePageMutationResponse {
    pageBuilder: {
        deletePage: {
            data: DeletePageMutationResponseData | null;
            error: PbErrorResponse | null;
        };
    };
}
export interface DeletePageMutationVariables {
    id: string;
}
export const DELETE_PAGE = gql`
    mutation PbDeletePage($id: ID!) {
        pageBuilder {
            deletePage(id: $id) {
                data {
                    latestPage {
                        id
                        status
                        version
                    }
                }
                ${error}
            }
        }
    }
`;

const PAGE_ELEMENT_FIELDS = /*GraphQL*/ `
    {
        id
        name
        type
        category
        content
        preview
    }
`;
/**
 * ##############################
 * List Page Elements Query
 */
export interface ListPageElementsQueryResponseDataPreview {
    src: string;
    meta: {
        width: number;
        height: number;
        aspectRatio: number;
    };
}
export interface ListPageElementsQueryResponseData {
    id: string;
    name: string;
    category: string;
    type: string;
    content: PbElement;
    preview: ListPageElementsQueryResponseDataPreview;
}

export interface ListPageElementsQueryResponse {
    pageBuilder: {
        data?: ListPageElementsQueryResponseData[];
        error?: PbErrorResponse;
    };
}
export const LIST_PAGE_ELEMENTS = gql`
    query PbListPageElements {
        pageBuilder {
            listPageElements {
                data ${PAGE_ELEMENT_FIELDS}
            }
        }
    }
`;

export const CREATE_PAGE_ELEMENT = gql`
    mutation PbCreatePageElement($data: PbCreatePageElementInput!) {
        pageBuilder {
            createPageElement(data: $data) {
                data ${PAGE_ELEMENT_FIELDS}
                ${error}
            }
        }
    }
`;

export const UPDATE_PAGE_ELEMENT = gql`
    mutation PbUpdatePageElement($id: ID!, $data: PbUpdatePageElementInput!) {
        pageBuilder {
            updatePageElement(id: $id, data: $data) {
                data ${PAGE_ELEMENT_FIELDS}
                ${error}
            }
        }
    }
`;
